/*
	File:		BRDF.cc

	Function:	

	Author:		Andrew Willmott

	Notes:		
*/

#include <complex.h>
#include "BRDF.h"

Phong::Phong(GCLReal rd, GCLReal rs, GCLReal rn, Colour rdClr, Colour rsClr) :
	rd(rd), rs(rs), rn(rn), rdClr(rdClr), rsClr(rsClr)
{
}

Colour Phong::BRDF(const Vector &normal, const Vector &light, const Vector &view)
{
	Colour	result;
	Vector	refl;
	
	result = rd * rdClr * Max(dot(light, normal), 0.0);
	refl = -light + 2.0 * dot(light, normal) * normal;
	result += rs * rsClr * pow(Max(dot(refl, view), 0.0), rn);

	return(result);
}

PhongH::PhongH(GCLReal rd, GCLReal rs, GCLReal rn, Colour rdClr, Colour rsClr) :
	rd(rd), rs(rs), rn(rn), rdClr(rdClr), rsClr(rsClr)
{
}

Colour PhongH::BRDF(const Vector &normal, const Vector &light, const Vector &view)
{
	Colour	result;
	Vector	halfway;
	
	halfway = light + view;
	normalise(halfway);

	result = rd * rdClr * Max(dot(light, normal), 0.0);
	result += rs * rsClr * pow(Max(dot(halfway, normal), 0.0), rn);

	return(result);
}

// Arguments:
// --- What we calculate ---
//   s*Rs + d*Rd
// --- Passed as Params ---
//   m = RMS slope of facets (a higher slope signifies a rougher surface)
//   d = coefficient of diffuse reflectance compared to specular
//   n = index of refraction of material
//   k = extinction coefficient (not currently used)
//   Rd_red, Rd_blue, Rd_green = Diffuse reflectance (surface color)
// --- Passed by the call to brdf ---
//   N = surface Normal
//   L = vector to Light source
//   V = vector to Viewer
//   wavlen = lambda = wavelength of incoming EM radiation (like, say, light)
// --- Implicit in other information ---
//   s = 1.0 - d = coefficient of specular reflectance
//   G = Geometric facet self-shadowing function, function of N, L, and V
//   D = surface distribution function, function of m, N, L, and V
//   F = fresnel function, function of n, V, and L
//   Rs = specular reflectance, function of N, L, V, F, D, and G



CookTorrance::CookTorrance(GCLReal m, GCLReal d, GCLReal n, Colour rdClr) :
	m(m), d(d), n(n), rdClr(rdClr)
{
}

Colour CookTorrance::BRDF(const Vector &N, const Vector &L, const Vector &V)
{
	Colour	rho(0.0, 0.0, 0.0);
	Vector	H;

	double N_dot_V = dot(N, V);
	if (N_dot_V <= 0) 
		return(rho);
	double N_dot_L = dot(N, L);
	if (N_dot_L <= 0)
		return(rho);

	double s = 1.0 - d;

	/* Precalculation for finding G */
	H = V;
	H += L; 	/* H = V + L */
	normalise(H);

	double N_dot_H = dot(N, H);
	double V_dot_H = dot(V, H);

	/* Calculate geometric self-shadowing coefficient, min of 3 possibilities */
	double G = Min(Min(1.0, (2.0 * N_dot_H * N_dot_V) / V_dot_H),
	               (2.0 * N_dot_H * N_dot_L) / V_dot_H);

	/* Precalculation for finding D */
	double N_dot_H_2 = N_dot_H * N_dot_H;
	double sin2_alpha = 1.0 - N_dot_H_2; 	/* sin^2(a) = 1.0 - cos^2(a) */

	/* tan^2(a) = sin^2(a) / cos^2(a) */
	double tan2_alpha = sin2_alpha / N_dot_H_2;

	/* Calculate Beckmann distribution function */
	double D = exp(-tan2_alpha / sqr(m))
	           / (sqr(m) * N_dot_H_2 * N_dot_H_2);

	/* Precalculation for F */
	double c = V_dot_H;
	double g = sqrt(sqr(n) + sqr(c) - 1.0);
	double temp_term = sqr(g - c) / sqr(g + c);
	double tt2 = (c * (g + c) - 1.0) / (c * (g - c) + 1.0);

	/* Calculate F.	This Fresnel function is valid only for k = 0, but it's
	apparently pretty close no matter what k is, within physical reason. */
	double F = 0.5 * temp_term * (1 + sqr(tt2));

	//	 Rs = specular reflectance, function of N, L, V, F, D, and G
	double Rs = (F / vl_pi) * (D / N_dot_L) * (G / N_dot_V);

	rho = d * rdClr + cWhite * s * Rs;
	rho *= N_dot_V;

	return(rho);
}

// --- He BRDF ----------------------------------------------------------------


#ifndef __SGI__

// This is, to say the least, somewhat more complicated than the others.

const Vector	lambdaRGB(.67, .55, .43);
const GCLReal	z0err = 0.0001;			// accepted error in value of z0

HeBRDF::HeBRDF(GCLReal tau, GCLReal sigma0, GCLReal d, GCLReal n_r, GCLReal n_i) :
	tau(tau), sigma0(sigma0), d(d), n_r(n_r), n_i(n_i)
{
}

Colour HeBRDF::BRDF(const Vector &n, const Vector &i, const Vector &o)
{
	GCLReal	result;
	
	lambda = lambdaRGB[0];
	epsilon = 1e-6;
	
	result = rhoDD(-norm(i), norm(o), n) + rhoUD(-norm(i), norm(o), n);

	return(result * cGreen);
}

GCLReal HeBRDF::shadowf(GCLReal t)
//	Shadowing function: Eqn (23)
{
	if (sigma0 == 0)
		return(1.0);
	else if (abs(t) < 0.0001)
		return(1.0);
	else if (abs(t) > (vl_halfPi - 0.0001))
		return(0.0);
	else
	{
		GCLReal	a = tau / (2.0 * sigma0 * tan(t));
		GCLReal	b = erfc(a);
		
		return(
			(1 - 0.5 * b) / 
			((exp(-sqr(a)) / (a * sqrt(vl_pi)) - b) / 2.0 + 1.0)
		);
	}
}

GCLReal HeBRDF::kK(GCLReal t)
//	Eqns (83), (84)
{
	if (abs(t) < epsilon)
		return(0);
	else
		return(
			tan(t) * erfc(tau / (2.0 * sigma0 * tan(t)))
		);
}

GCLReal HeBRDF::gFunc(GCLReal cti, GCLReal ctr)
//	g; Eqn (79)
{
	GCLReal sigmaIR = sigma(acos(cti), acos(ctr));
	
//	DUMP("sigma", sigmaIR);
	
	return(
		sqr(2 * vl_pi / lambda * sigmaIR * (cti + ctr))
	);
}

GCLReal myDsum(GCLReal tf, GCLReal gt)
{
	/*
		Purpose: form

			sum 1..infinity: term(i)
		
			term(i) = pow(gt, i) / Real(fact(i) * i)
				 * exp(tf / Real(i)) * exp(-gt);
			 
		Better than this would be to start with term i = gt, and
		work outwards. (The max in gpowfact will occur around this
		term in the series.)
		
	*/

	Int		i;
	GCLReal	sum = 0, term, lastTerm = 0;
	GCLReal	gpowfact;
	
	gpowfact = gt;
	
	for (i = 1; i < 400; i++)
	{
		term = gpowfact * exp(-gt + tf / GCLReal(i));
		// sometimes get underflow here
//		if (IsNANorINF(term))
//		    return sum;
		gpowfact *= gt * GCLReal(i) / sqr(GCLReal(i + 1));
		 
		sum += term;

		//	If we're over the hump of term(i), and the term
		//	is small enough, bail.

		if (term < 0.000001 && term < lastTerm)
			break;
	
		lastTerm = term;	
	}

#ifdef HE_DEBUG	
	cout << "*** sum = " << sum << endl;
#endif	
	
	return(sum);
}

GCLReal HeBRDF::kD(const Vector &ki, const Vector &kr, const Vector &ns)
//	Distribution function D, eqn (78)
{
	GCLReal gt, mf, tf, cti, ctr;
	Vector v;
	
	cti = -dot(ki, ns);
	ctr =  dot(kr, ns);
	v = kr - ki;
		
	gt = gFunc(cti, ctr);	
	mf = sqr(vl_pi) * sqr(tau) / (4.0 * sqr(lambda));
	tf = sqr(2.0 * vl_pi / lambda) *
		-(sqrlen(v) - sqr(dot(ns, v))) * sqr(tau) / 4.0;

#if defined(HE_DEBUG) && 1
	
	DUMP("cti", cti);
	DUMP("ctr", ctr);
	cout << "gt = " << gt << endl;
	cout << "mf = " << mf << endl;
	cout << "tf = " << tf << endl;
#endif

	return(mf * myDsum(tf, gt));
}

GCLReal HeBRDF::kG(const Vector &ki, const Vector &kr, const Vector &ns)
// Geometrical factor G, eqn (76)
{
#ifdef PARANOID
	Vector	si, sr, kc, pi, pr;
	GCLReal	skc;
	
	si = norm(cross(ki, ns));
	sr = norm(cross(kr, ns));
	pi = cross(si, ki);
	pr = cross(sr, kr);
	kc = cross(kr, ki);
	skc = sqrlen(kc);
	GCLReal top =
	    (sqr(dot(sr, ki)) + sqr(dot(pr, ki))) *
	    (sqr(dot(si, kr)) + sqr(dot(pi, kr)));
	Assert(abs(top - sqr(skc)) < 1e-4, "we were wrong\n");

	// NOTE! last three factors of He's formula 76 multiplied together
	// equal 1!
	// We use the simplified formula below.
#endif
	
	Vector v = kr - ki;
	
	return(sqr(sqrlen(v) / dot(ns, v)));
}


/*
	Code for finding z0,  the root to the following equation:

	sqrt(vl_halfPi) * z0 == 
		sigma0 / 4.0 * (kK(ti) + kK(tr)) * exp(-sqr(z0 / sigma0) / 2.0)
*/

GCLReal HeBRDF::z0d(GCLReal Ki, GCLReal z)
{
	return(
		-Ki / (4.0 * sigma0) * z * exp(-sqr(z / sigma0) / 2.0) - sqrt(vl_halfPi)
	);
}

GCLReal HeBRDF::z0lim(GCLReal x)
{
	if (x > 0)
		return(Max(x, z0err));
	else
		return(Min(x, -z0err));
}

GCLReal HeBRDF::z0off(GCLReal Ki, GCLReal z)
{
	return(
		(sigma0 / 4.0 * Ki * exp(-sqr(z / sigma0) / 2) - sqrt(vl_halfPi) * z) /
			z0lim(z0d(Ki, z))
	);
}

GCLReal HeBRDF::z0rootf(GCLReal Ki, GCLReal x0, GCLReal x1, Int i)
{
	if (i <= 0)
		return(0);
	else if (z0err > abs(x1 - x0))
		return(x1);
	else
		return(z0rootf(Ki, x1, x1 - z0off(Ki, x1), i - 1));
}

GCLReal HeBRDF::z0root(GCLReal ti, GCLReal tr)
{
	if (abs(ti) < 0.6 && abs(tr) < 0.6)
		return(0.0);
	else if (abs(ti) >= (vl_halfPi - 0.0001) || abs(tr) >= (vl_halfPi - 0.0001))
		return(1.0);
	else
	 	return(z0rootf(kK(ti) + kK(tr), 0.1, -z0off(kK(ti) + kK(tr), 0.1), 100));
}

GCLReal HeBRDF::sigma(GCLReal ti, GCLReal tr)
//	Roughness factor sigma; eqn (80)
{
	if (abs(sigma0) < epsilon)
		return(sigma0);
	else
	{
		GCLReal z0 = z0root(ti, tr);
		GCLReal lhs, rhs;

		lhs = sqrt(vl_halfPi) * z0;
		rhs = sigma0 / 4.0 * (kK(ti) + kK(tr)) * exp(-sqr(z0 / sigma0) / 2.0);

#ifdef HE_DEBUG
		cout << "ti = " << ti << endl;
		cout << "tr = " << tr << endl;
		cout << "z0 = " << z0 << endl;
		cout << "lhs = " << lhs << endl;
		cout << "rhs = " << rhs << endl;
		cout << "kti = " << kK(ti) << endl;
		cout << "ktr = " << kK(tr) << endl;
#endif
		
		return(sigma0 / sqrt(1 + sqr(z0 / sigma0)));
	}
}

// I think double_complex is the most recent way of doing it,
// but the irix headers still need the template version
// XXX look this up sometime
typedef complex<double> double_complex;

GCLReal fresnel(GCLReal ct, GCLReal nr, GCLReal ni)
//	The fresnel factor! 
{
	double_complex n(nr, ni);
	double_complex	ct1(ct, 0.0);
	double_complex	ct2 = sqrt(1.0 - (1.0 - sqr(ct1)) / sqr(n));
	double_complex a, b;
	
	a = (n * ct1 - ct2) / (n * ct1 + ct2);
	b = (n / ct1 - 1.0 / ct2) / (n / ct1 + 1.0 / ct2);
	
	return((sqr(real(a)) + sqr(imag(a)) + sqr(real(b)) + sqr(imag(b))) / 2.0);
}


GCLReal HeBRDF::rhoSp(const Vector &ki, const Vector &kr, const Vector &ns)
// Eqn (70)
{
	GCLReal fr, sti, str, cti, ctr;
	
	fr = fresnel(len(kr - ki) / 2.0, n_r, n_i);
	cti = -dot(ki, ns);
	ctr =  dot(kr, ns);
	sti = shadowf(acos(cti));
	str = shadowf(acos(ctr));

	return(
		exp(-gFunc(cti, ctr)) * sti * str * fr / cti
	);
}

GCLReal HeBRDF::rhoDD(const Vector &ki, const Vector &kr, const Vector &ns)
{
	GCLReal ti, tr, cti, ctr;
	GCLReal fr, kg, kd, sti, str;
	
#ifdef HE_DEBUG
	cout << "ki = " << ki << endl;
	cout << "kr = " << kr << endl;
	cout << "ns = " << ns << endl;
#endif

	cti = -dot(ki, ns);
	ctr =  dot(kr, ns);
	ti = acos(cti);
	tr = acos(ctr);

	fr = fresnel(len(kr - ki) / 2.0, n_r, n_i);
	kg = kG(ki, kr, ns);
	kd = kD(ki, kr, ns);
	sti = shadowf(ti);
	if (sti < epsilon)
		return(0.0);
	str = shadowf(tr);
	if (str < epsilon)
		return(0.0);
	
#ifdef HE_DEBUG
	cout << "ti = " << ti << endl;
	cout << "tr = " << tr << endl;
	cout << "fr = " << fr << endl;
	cout << "kg = " << kg << endl;
	cout << "kd = " << kd << endl;
	cout << "sti = " << sti << endl;
	cout << "str = " << str << endl;
#endif

//	Eqn (72)

	GCLReal result = 
		fr / vl_pi *
		kd * sti * str * kg /
		(cti * ctr);
		
	return(result);
}

GCLReal HeBRDF::rhoUD(const Vector &ki, const Vector &/*kr*/, const Vector &ns)
{
	GCLReal result = -dot(ki, ns) * d;
	
	return(result);
}

#ifdef HE_DEBUG

void main()
{
	HeBRDF	he;
	GCLReal	x, result;
	
#if 0
	// Ceramic

	he.param[0].val = 1.5;
	he.param[1].val = 0;
	he.param[2].val = 1;
	he.param[3].val = 1.45;
#endif
#if 1
	// Aluminium
	
	he.param[0].val = 0.770058;
	he.param[1].val = 6.08351;
	he.param[2].val = 1.77;
	he.param[3].val = 0.28;
#endif

	he.lambda = 0.5;

#if 1
	for (x = 0.1; x < vl_pi; x += 0.1)
	{
		result = he.rhoDD(norm(Vector(1, -1, 0)), norm(Vector(cos(x), sin(x), 0)), Vector(0, 1, 0));	
		cout << "*** " << x << " = " << result << endl;
	}
#else
	for (x = 0.1; x < 1.0; x += 0.1)
	{
		result = he.gFunc(x, 0.4);
		cout << "*** " << x << " = " << result << endl;
	}
	x = x / 0.0;
#endif
}
#endif

#endif
