/*
	File:			Colour.h
	
	Function:		Definitions and classes for handling Colour

	Author(s):		Andrew Willmott

	Copyright:		(c) 1995-2000, Andrew Willmott
 */
 
#ifndef __Colour__
#define __Colour__

#include "GCLConfig.h"

#ifndef GCL_CLR_DOUBLE
#include "vl/VLf.h"
typedef Vec3f Colour;
typedef Vec4f Colour4;
typedef Mat3f ClrMat;
typedef Mat4f ClrTrans;
typedef Float ClrReal;
#else
#include "vl/VLd.h"
typedef Vec3d Colour;
typedef Vec4d Colour4;
typedef Mat3d ClrMat;
typedef Mat4d ClrTrans;
typedef Double ClrReal;
#endif

#include "cl/Array.h"
#include "NArray.h"

/**	\typedef Colour
	A standard tri-stimulus colour.
	We don't store a tag saying what basis functions the colour
	depends on for efficiency reasons. 
*/
/** \typedef Colour4
	A Colour with a transparency channel. 
*/

inline Colour	RGBCol(ClrReal r, ClrReal g, ClrReal b)
				{ return(Colour(r, g, b)); }
Colour			HSVCol(ClrReal hue, ClrReal saturation, ClrReal value);

typedef Colour Reflectance;
typedef NArray<Colour> ColourList;
typedef NArray<Colour4> Colour4List;
typedef ColourList ReflList;

Colour &ClipColour(Colour &c);
///< Clips colour to [0, 1] for display
Colour &ClipColourZero(Colour &c);
///< Clips colour to zero only.
inline Colour ClrMix(const Colour &c1, const Colour &c2, ClrReal mix)
{
	return(c1 * (1.0 - mix) + c2 * mix);
}
///< Mixes two colours


// --- Useful conversion factors/functions ------------------------------------

const Reflectance cRGBToNTSCLum(0.299, 0.587, 0.114);	
/**< Converts RGB to greyscale, assuming we're in an NTSC colourspace with
	gamma = 2.2	*/

const Reflectance cRGBToLum(0.3086, 0.6094, 0.0820);	
///< Convert RGB to greyscale, assuming a linear colour space.

const ClrMat cNTSCToCIE(
				0.67, 0.21, 0.14,			// From Sillion & Puesch
				0.33, 0.71, 0.08,			// "Radiosity", pp 211.
				0.00, 0.08, 0.78); 

const ClrMat cCIEToNTSC(
				1.730, -0.482, -0.261,
			   -0.814,  1.652, -0.023,
				0.083, -0.169,  1.284); 

const ClrMat cRGBToYIQ(
				0.299,  0.587,  0.114,
				0.596, -0.275, -0.321,
				0.212, -0.523,  0.311);

const ClrMat cYIQToRGB(
				1.0,   0.956,   0.621,
				1.0,  -0.272,  -0.647,
				1.0,  -1.105,   1.702);

Colour YUVToRGB(const Colour &yuv);
Colour RGBToYUV(const Colour &rgb);
///< YUV == YCbCr


// --- Useful colour constants ------------------------------------------------

// RGB Constants

const Colour cBlack(0, 0, 0);
const Colour cWhite(1, 1, 1);

const Colour cGrey(cWhite * 0.5);
const Colour cGrey05(cWhite * 0.05);
const Colour cGrey10(cWhite * 0.1);
const Colour cGrey25(cWhite * 0.25);
const Colour cGrey50(cWhite * 0.5);
const Colour cGrey75(cWhite * 0.75);
const Colour cGrey90(cWhite * 0.90);
const Colour cGrey95(cWhite * 0.95);

const Colour cRed(1, 0, 0);
const Colour cOrange(1, 0.5, 0);
const Colour cYellow(1, 1, 0);
const Colour cGreen(0, 1, 0);
const Colour cCyan(0, 1, 1);
const Colour cBlue(0, 0, 1);
const Colour cPurple(1, 0, 1);

// HSV constants

const Int hsvRed    =    0;
const Int hsvOrange =   30;
const Int hsvYellow =   60;
const Int hsvGreen  =  120;
const Int hsvCyan   =  180;
const Int hsvBlue   =  240;
const Int hsvPurple =  300;


// --- Colour transformations -------------------------------------------------

/// use Image::Transform() to apply these to an image.

ClrTrans	RGBScale(ClrReal rscale, ClrReal gscale, ClrReal bscale);
ClrTrans	RGBOffset(const Colour &c);
ClrTrans 	RGBReplace(const Colour &c);
ClrTrans	RGBToLum();

ClrTrans	RGBSaturate(ClrReal sat);
			/**< Saturate image
				sat < 1.0 decreases saturation towards 0.0 (greyscale),
				sat > 1.0 increases it.	
			*/
ClrTrans	RGBPixelMix(ClrReal mix, const Colour &c);
			/**< Mix with colour
				If you use c = cBlack, mix < 1.0 will darken the image,
				and mix > 1.0 will lighten it.
				If you use the average colour of an image, c = img.AverageColour(),
				mix < 1.0 will lower contrast, mix > 1.0 raise it.
			*/
ClrTrans	RGBHueRotate(ClrReal degrees);
			///< Rotate hue of image in HSV space.
#endif
