/*
	File:			Image.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott
*/

#include "Image.h"
#include "GCLConfig.h"
#include "VecUtil.h"

#include <iostream.h>
#include <stdio.h>
#include <ctype.h>


// --- Image class ------------------------------------------------------------



Int Image::sJPEGQuality = 90;

Void Image::Clear(const Colour &c)
// should almost always be overridden
{
	Int			x, y;

	for (y = 0; y < height; y++)
		for (x = 0; x < width; x++)
			SetPixel(x, y, c);
}

Void Image::CopyFrom(Int x, Int y, Image &img)
{
	Int		test = img.Width();
	Int		copyWidth = Min(img.Width(), width - x);
	Int 	copyHeight = Min(img.Height(), height - y);
	Colour 	*buffer = new Colour[copyWidth];
	Int 	i;
	
	for (i = 0; i < copyHeight; i++)
	{
		img.GetSpan(i, x, copyWidth, buffer);
		SetSpan(y++, 0, copyWidth, buffer);
	} 
	
	delete[] buffer;
}

Void Image::GammaCorrect(ClrReal gamma)
{
	ClrReal		    invGamma = 1.0 / gamma;
	Colour          c;
	Int				x, y;
	
	for (y = 0; y < Height(); y++)
		for (x = 0; x < Width(); x++)
		{
			c = GetPixel(x, y);

			c[0] = pow(c[0], invGamma);
			c[1] = pow(c[1], invGamma);
			c[2] = pow(c[2], invGamma);

			SetPixel(x, y, c);
		}
}

Void Image::Scale(ClrReal scale)
{
	Colour 	*buffer = new Colour[width];
	Int 	i, j;
	Colour	result;
	
	for (i = 0; i < height; i++)
	{
		GetSpan(i, 0, width, buffer);
		for (j = 0; j < width; j++)
			buffer[j] *= scale;
		SetSpan(i, 0, width, buffer);
	} 
}

ClrReal Image::MaxComponent()
{
	Int			x, y;
	ClrReal		maxCmpt = 0.0;
	Colour		c;
	
	for (y = 0; y < height; y++)
		for (x = 0; x < width; x++)
		{
			c = GetPixel(x, y);
			maxCmpt = Max(maxCmpt, MaxCmpt(c));
		}

	return(maxCmpt);
}

Colour Image::AverageColour()
{
	Colour 	*buffer = new Colour[width];
	Int 	i, j;
	Colour	result = cBlack;
	
	for (i = 0; i < height; i++)
	{
		Colour	bsum = cBlack;
		
		GetSpan(i, 0, width, buffer);
		for (j = 0; j < width; j++)
			bsum += buffer[j];
		bsum /= width;
		result += bsum;
	}

	delete[] buffer;
	result /= height;

	return(result);
}

Void Image::FindComponentBounds(Colour &min, Colour &max)
{
	Int			x, y;
	Colour		c;
	
	for (y = 0; y < height; y++)
		for (x = 0; x < width; x++)
		{
			c = GetPixel(x, y);
			FindMinCmpts(min, c, min);
			FindMaxCmpts(max, c, max);
		}
}

Void Image::Transform(const ClrTrans &trans)
{
	Colour 	*buffer = new Colour[width];
	Int 	i, j;
	
	for (i = 0; i < height; i++)
	{
		GetSpan(i, 0, width, buffer);
		for (j = 0; j < width; j++)
		{
			buffer[j] = xform(trans, buffer[j]);
			ClipColour(buffer[j]);
		}
		SetSpan(i, 0, width, buffer);
	} 

	delete[] buffer;
}

Void Image::Over(Int x, Int y, Image &img)
{
	Int		test = img.Width();
	Int		copyWidth = Min(img.Width(), width - x);
	Int 	copyHeight = Min(img.Height(), height - y);
	Colour 	*buffer = new Colour[copyWidth];
	Int 	i;

	// XXX UNFINISHED: want to do composite operation: d = d * (1 - s.alpha) + s
	
	for (i = 0; i < copyHeight; i++)
	{
		img.GetSpan(i, x, copyWidth, buffer);
		SetSpan(y++, 0, copyWidth, buffer);
	} 
	
	delete[] buffer;
}

Void Image::DownSample(Image &out)
/// quick hack for now. do proper arbitrary filtered scaling sometime.
{
	Int		i, j;
	static ColourList span1, span2, outSpan;

	out.SetSize(Width() / 2, Height() / 2);	
	span1.SetSize(Width());
	span2.SetSize(Width());
	outSpan.SetSize(out.Width());

	for (i = 0; i < out.Height(); i++)
	{
		GetSpan(2 * i, 0, Width(), span1.Ref());
		GetSpan(2 * i + 1, 0, Width(), span2.Ref());

		for (j = 0; j < out.Width(); j++)
		{
			outSpan[j]  = span1[2 * j] + span1[2 * j + 1];
			outSpan[j] += span2[2 * j] + span2[2 * j + 1];
			outSpan[j] /= 4.0;
		}

		out.SetSpan(i, 0, out.Width(), outSpan.Ref());
	}
}


// These should almost always be overridden with more efficient routines

Void Image::SetRealPixel(Int x, Int y, ClrReal r, ImgChannel channel)
{
	if (channel == chMono)
		SetPixel(x, y, cWhite * r);
	else if (channel <= chBlue)
	{
		Colour c = GetPixel(x, y);

		c[channel - chRed] = r;		
		SetPixel(x, y, c);
	}
}

ClrReal	Image::GetRealPixel(Int x, Int y, ImgChannel channel) const
{
	if (channel <= chBlue)
	{
		Colour c = GetPixel(x, y);

		if (channel == chMono)
			return(dot(cWhite, c) * (1.0 / 3.0));
		else
			return(c[channel - chRed]);
	}
	else
		return(0.0);
}

Void Image::SetRealSpan(Int row, Int start, Int length, 
						const ClrReal *src, ImgChannel channel)
{
	Int		i;

	for (i = 0; i < length; i++)
		SetRealPixel(start + i, row, src[i], channel);
}

Void Image::GetRealSpan(Int row, Int start, Int length, 
						ClrReal *dst, ImgChannel channel) const
{
	Int		i;

	for (i = 0; i < length; i++)
		dst[i] = GetRealPixel(start + i, row, channel);
}

Void Image::SetRGBASpan(Int row, Int start, Int length, const RGBAPixel *src) 
{
	Int		i;
	Colour	*buffer = new Colour[length];

	for (i = 0; i < length; i++)
		buffer[i] = RGBAToColour(src[i]);

	SetSpan(row, start, length, buffer);

	delete[] buffer;
}

Void Image::GetRGBASpan(Int row, Int start, Int length, RGBAPixel *dst) const
{
	Int		i;
	Colour	*buffer = new Colour[length];

	GetSpan(row, start, length, buffer);

	for (i = 0; i < length; i++)
		dst[i] = ColourToRGBA(buffer[i]);

	delete[] buffer;
}

Void Image::SetByteSpan(Int row, Int start, Int length, 
						const Byte *src, ImgChannel channel)
{
	Int		i;
	Colour	*buffer = new Colour[length];

	for (i = 0; i < length; i++)
		buffer[i] = ByteToColour(src[i]);

	SetSpan(row, start, length, buffer);

	delete[] buffer;
}

Void Image::GetByteSpan(Int row, Int start, Int length, 
						Byte *dst, ImgChannel channel) const
{
	Int		i;
	Colour	*buffer = new Colour[length];

	GetSpan(row, start, length, buffer);

	for (i = 0; i < length; i++)
		dst[i] = ColourToByte(buffer[i]);

	delete[] buffer;
}


// --- File methods -----------------------------------------------------------

Char *tImageFormats[] = 
{
	"ppm",		"PPM image format",
#ifdef GCL_TIFF
	"tif",		"Adobe TIFF format",
#endif
#ifdef GCL_JPEG
	"jpg",		"JPEG format",
#endif
#ifdef GCL_GIF
	"gif",		"GIF format (read only)",
#endif
	0
};

Void Image::PrintSupportedFormats(ostream &s)
{
	Char	**p = tImageFormats;
	
	s << "Supported image file extensions: " << endl;

	while (*p)
	{
		s << String().Printf("%-5s %s\n", *p, *(p + 1));
		p += 2;
	}
}

Int Image::Save(FileName &filename)
{
	if (filename.GetExtension() == "ppm")
		return(SavePPM(filename.GetPath()));
#ifdef GCL_TIFF
	else if (filename.GetExtension() == "tiff" || filename.GetExtension() == "tif")
		return(SaveTIFF(filename.GetPath()));
#endif
#ifdef GCL_JPEG
	else if (filename.GetExtension() == "jpg")
		return(SaveJPEG(filename.GetPath()));
#endif
	else
	{
#ifdef GCL_TIFF
        if (filename.GetExtension() != "")
			cerr << "Unknown image file extension:"
				 << " saving in TIFF format" << endl;
		filename.SetExtension("tif");
		return(SaveTIFF(filename.GetPath()));
#else
        if (filename.GetExtension() != "")
			cerr << "Unknown image file extension:"
				 << " saving in PPM format" << endl;
		filename.SetExtension("ppm");
		return(SavePPM(filename.GetPath()));
#endif		
	}
}

static StrConst kImageExtensions[] =
{
	"ppm",
	"pgm",
#ifdef GCL_TIFF
	"tif",
	"tiff",
#endif
#ifdef GCL_JPEG
	"jpg",
#endif
#ifdef GCL_GIF
	"gif",
#endif
	0
};

Int Image::Load(FileName &filename)
{
	Int		result = filename.FindFileExtension(kImageExtensions);

	if (result == kBadExtension)
		cerr << "error: unknown image extension '" << filename.GetExtension()
			 << "'" << endl;
	else if (result == kFileNotFound)
		cerr << "error: can't find image file " << filename.GetPath() << endl;

	if (result < 0)
		return(result);

	filename.DecompressSetup();
	if (result < 2)
		return(LoadPPM(filename.GetPath()));
#ifdef GCL_TIFF
	else if (filename.GetExtension() == "tiff" || filename.GetExtension() == "tif")
		return(LoadTIFF(filename.GetPath()));
#endif
#ifdef GCL_JPEG
	else if (filename.GetExtension() == "jpg")
		return(LoadJPEG(filename.GetPath()));
#endif
#ifdef GCL_GIF
	else if (filename.GetExtension() == "gif")
		return(LoadGIF(filename.GetPath()));
#endif
	filename.DecompressCleanup();

	_Error("Unhandled extension");

	return(-1);
}


// --- PPM read/write ---------------------------------------------------------


Int Image::SavePPM(const Char *filename)
{
    int			i, j;
	FILE		*ppm;
	RGBAPixel	*p, *buffer = new RGBAPixel[width];
	
	if (!buffer)
		return(-1);
	ppm = fopen(filename, "w");
	if (!ppm)
		return(-1);
	
	fprintf(ppm, "P6\n%u %u\n255\n", width, height);

    for (j = height - 1; j >= 0; j--)
    {
    	GetRGBASpan(j, 0, width, buffer);
    	p = buffer;
        for (i = 0; i < width; i++, p++)
 		{
            fputc(p->ch[rgba_R], ppm);
            fputc(p->ch[rgba_G], ppm);
            fputc(p->ch[rgba_B], ppm);
        }
    }

	delete[] buffer;
    return(fclose(ppm));
}

static Int ppmReadInt(FILE *ppm)
{
	Int result, c;

	// chomp space & comments
	c = fgetc(ppm);

	while(isspace(c) || c == '#')
	{
		if (c == '#')
			do 				
				c = fgetc(ppm);
			while (c != '\n' && c != '\r' && c != EOF);
		
		c = fgetc(ppm);
	}

	ungetc(c, ppm);
	fscanf(ppm, "%d", &result);

	return(result); 
}

Int Image::LoadPPM(const Char *filename)
{
    Int			i, j;
	Int			x, y, level;
	Char		p1, p2;
	Bool		isMono;
	FILE		*ppm;
	RGBAPixel	*p, *buffer;
	Byte		*mp, *mBuffer;
	
	ppm = fopen(filename, "r");
	if (!ppm)
		return(-1);

	p1 = fgetc(ppm);
	p2 = fgetc(ppm);

	if (p1 != 'P' || p2 <= '1' || p2 > '6')
	{
		_Warning("(LoadPPM) not a PPM file");
		return(-1);
	}

	if (p2 < '2' || p2 == '4' || p2 > '6')
	{
		_Warning("(LoadPPM) Only handles PGM and PPM files");
		return(-1);
	}

	isMono = (p2 == '2' || p2 == '5');
	
	x = ppmReadInt(ppm);
	y = ppmReadInt(ppm);
	level = ppmReadInt(ppm);

	if (level != 255)
		_Warning("(LoadPPM) Ignoring scaling factor.");
	
	SetSize(x, y);

	if (fgetc(ppm) != '\n')
	{
		_Warning("(LoadPPM) Corrupt PPM file");
		return(-1);
	}

	if (isMono)
		mBuffer = new Byte[width];
	else
		buffer = new RGBAPixel[width];

    for (j = height - 1; j >= 0; j--)
    {
		if (p2 == '2')
			for (i = 0, mp = mBuffer; i < width; i++, mp++)
			{
				Int		b;

				fscanf(ppm, "%d", &b);
				*mp = b;
			}
		else if (p2 == '3')
			for (i = 0, p = buffer; i < width; i++, p++)
			{
				Int		b;
				fscanf(ppm, "%d", &b);
				p->ch[rgba_R] = b;
				fscanf(ppm, "%d", &b);
				p->ch[rgba_G] = b;
				fscanf(ppm, "%d", &b);
				p->ch[rgba_B] = b;
				p->ch[rgba_A] = 255;
			}
		else if (p2 == '5')
			fread(mBuffer, 1, width, ppm);
		else if (p2 == '6')
			for (i = 0, p = buffer; i < width; i++, p++)
			{
				// UUU #ifdef endianess here sometime
				p->ch[rgba_R] = fgetc(ppm);
				p->ch[rgba_G] = fgetc(ppm);
				p->ch[rgba_B] = fgetc(ppm);
				p->ch[rgba_A] = 255;
			}

        if (isMono)
        	SetByteSpan(j, 0, width, mBuffer);
        else
        	SetRGBASpan(j, 0, width, buffer);
	}

	if (isMono)
		delete[] mBuffer;
	else
		delete[] buffer;

	return(fclose(ppm));
}


// --- TIFF read/write --------------------------------------------------------


#ifdef GCL_TIFF

#include "tiffio.h"

Int Image::SaveTIFF(const Char *filename)
{
    TIFF	*tif;
    UInt32	samples_per_pixel = 3;
    UInt32	w = Width();
    UInt32	h = Height();
    UInt32	scanline_size = samples_per_pixel * w;
    UInt32	i, k;
	Int		y;
    Char	*scanline_buf;
	RGBAPixel *rgba_buf;

    tif = TIFFOpen(filename, "w");
    if (!tif)
		return(-1);

    TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, w);
    TIFFSetField(tif, TIFFTAG_IMAGELENGTH, h);

    /* These are the charateristics of our Pic data */
    TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_LEFTTOP);
    TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, samples_per_pixel);
    TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
#ifdef GCL_PATENTS_SUCK
	// use PKZIP compression to escape the pitiful unisys nonsense.
	// currently, RH 6.2's tiff libs print a pissy little message
	// LZW and don't use it, but they haven't enabled support
	// for deflate compression. I suppose that would have just
	// been too bright of them. If you want compressed images,
	// you'll have to rebuild with either LZW reenabled, or deflate
	// enabled.
	TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE); 	
#else
    TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
#endif
    
    if( TIFFScanlineSize(tif) != scanline_size )
    {
		fprintf(stderr,
			"TIFF: Mismatch with library's expected scanline size!\n");
		return(-1);
    }
    
    TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(tif, 0));

    scanline_buf = new Char[scanline_size];
    rgba_buf = new RGBAPixel[scanline_size];

    if (!scanline_buf)
	{
		fprintf(stderr, "TIFF: Can't allocate scanline buffer!\n");
		return(-1);
    }
	
    for (y = h - 1; y >= 0; y--)
    {
		GetRGBASpan(y, 0, w, rgba_buf);
		k = 0;
		for (i = 0; i < w; i++)
		{
			scanline_buf[k++] = rgba_buf[i].ch[rgba_R];
			scanline_buf[k++] = rgba_buf[i].ch[rgba_G];
			scanline_buf[k++] = rgba_buf[i].ch[rgba_B];
		}
		
		TIFFWriteScanline(tif, scanline_buf, h - y - 1, 0);
		/* note that TIFFWriteScanline modifies the buffer you pass it */
    }

    delete[] scanline_buf;
	delete[] rgba_buf;
	TIFFClose(tif);

    return(0);
}

Int Image::LoadTIFF(const Char *filename)
{
    TIFF			*tif;
	TIFFRGBAImage	img;
	Int				result = -1;
	Char			emsg[1024];
	
    tif = TIFFOpen(filename, "r");
    if (!tif)
	{
		_Warning("(LoadTIFF) Couldn't find file.");
		return(-1);
	}

	if (TIFFRGBAImageBegin(&img, tif, 0, emsg)) 
	{
		SetSize(img.width, img.height);

#ifdef GCL_BIG_END
		// TIFF unpacks data in AGBR format, which is endian dependent.
		// If we're on big-end machines, it matches with our format,
		// and we can write the data straight in, otherwise we have to
		// convert.
		if (Tag() == imgRGBATag)
			TIFFRGBAImageGet(&img, ((RGBAImage*) this)->Data(), 
							 img.width, img.height); 
		else
		{
#endif
			Int		npixels = img.width * img.height;
			UInt32	*raster = 
				(UInt32*) _TIFFmalloc(npixels * sizeof (UInt32));

			// should use TIFF put methods here, but they're too poorly
			// documented to waste time on just at the mo'
			if (raster) 
			{
				if (TIFFRGBAImageGet(&img, (uint32*) raster, img.width, img.height)) 
				{
					Int			i, j;
					RGBAPixel	*rgbaBuf = new RGBAPixel[img.width];

					if (rgbaBuf)
						for (i = 0; i < img.height; i++)
						{
							for (j = 0; j < img.width; j++)
							{
								UInt32	pixel = raster[i * img.width + j];

								rgbaBuf[j].ch[rgba_R] = TIFFGetR(pixel);
								rgbaBuf[j].ch[rgba_G] = TIFFGetG(pixel);
								rgbaBuf[j].ch[rgba_B] = TIFFGetB(pixel);
								rgbaBuf[j].ch[rgba_A] = TIFFGetA(pixel);
							}
							SetRGBASpan(i, 0, img.width, rgbaBuf);
						}

					delete[] rgbaBuf;
				}
				else
				{
					_Warning("(LoadTIFF) TIFFRGBAImageGet failed");
					return(-1);
				}
				_TIFFfree(raster);
			}
			else
			{
				_Warning("(LoadTIFF) couldn't allocate raster data");
				return(-1);
			}

#ifdef GCL_BIG_END
		}
#endif

		TIFFRGBAImageEnd(&img);
		result = 0;
	} 
	else
		TIFFError("(LoadTIFF)", emsg);


	TIFFClose(tif);

    return(result);
}

#endif


// --- GIF read/write --------------------------------------------------------

#ifdef GCL_GIF

// Bog off, UNISYS

extern "C"
{
#include "gif_lib.h"
}

// why on earth isn't this handled by DGifSlurp? 
// and why does that code assume char is unsigned? 
static Int tInterlacedOffset[] = { 0, 4, 2, 1 };
static Int tInterlacedJump[] = { 8, 8, 4, 2 };

Int Image::SaveGIF(const Char *filename)
{
	return(-1);
}

Int Image::LoadGIF(const Char *filename)
{
	GifFileType		*gifFile;
	SavedImage		*theGIFImg;
	ColorMapObject	*clrMap;
	RGBAPixel		*buffer;
	Int				i, j, result = -1;
	Int				pass, destRow, destStep;
	
	gifFile = DGifOpenFileName(filename);	
	if (gifFile)
	{
		if (DGifSlurp(gifFile) == GIF_OK)
		{
			buffer = new RGBAPixel[gifFile->SWidth];

			theGIFImg = gifFile->SavedImages;	// first image
			
			SetSize(theGIFImg->ImageDesc.Width, theGIFImg->ImageDesc.Height);

			if (theGIFImg->ImageDesc.ColorMap)
				clrMap = theGIFImg->ImageDesc.ColorMap;
			else if (gifFile->SColorMap)
				clrMap = gifFile->SColorMap;
			else
			{
				_Warning("GIF file contains no colormap!");
				return(-1);
			}
			
			if (theGIFImg->ImageDesc.Interlace)
			{
				destRow = tInterlacedOffset[0];
				destStep = tInterlacedJump[0];
				pass = 0;
			}
			else
			{
				destRow = 0;
				destStep = 1;
			}
				
			for (i = 0; i < theGIFImg->ImageDesc.Height; i++)
			{
				Byte *sp = (Byte*) theGIFImg->RasterBits + i * theGIFImg->ImageDesc.Width;

			    for (j = 0; j < theGIFImg->ImageDesc.Width; j++)
				{
					Byte idx = *sp++;

					buffer[j].ch[rgba_R] = clrMap->Colors[idx].Red;
					buffer[j].ch[rgba_G] = clrMap->Colors[idx].Green;
					buffer[j].ch[rgba_B] = clrMap->Colors[idx].Blue;
					buffer[j].ch[rgba_A] = 255;
				}
				
				SetRGBASpan(height - 1 - destRow, 0, theGIFImg->ImageDesc.Width, buffer);
				destRow += destStep;
				if (destRow >= height && theGIFImg->ImageDesc.Interlace)
				{
					pass++;
					Assert(pass < 4, "Too many passes in GIF file");
					destRow = tInterlacedOffset[pass];
					destStep = tInterlacedJump[pass];
				}
		    }
			
		    delete[] buffer;
			result = 0;
		}
		else
			PrintGifError();

		if (DGifCloseFile(gifFile) != GIF_OK)
			PrintGifError();
	}
	else
		PrintGifError();

	return(result);
}

#endif


// --- JPEG read/write --------------------------------------------------------

#ifdef GCL_JPEG

extern "C"
{
#include "jpeglib.h"
}

Int Image::SaveJPEG(const Char *filename)
{
	FILE                        *outfile;
	JSAMPROW                    rowPtr[1];
	struct jpeg_compress_struct compressObj;
	struct jpeg_error_mgr       jerr;
	JSAMPLE                     *scanlineBuf, *p;
	Int                         i, j;
	
	compressObj.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&compressObj);

	//	Set up file for output... 

	if ((outfile = fopen(filename, "wb")) == NULL)
	{
		_Warning("(Image::SaveJPEG) can't open for writing: " + StrConst(filename));
		return(-1);
	}
	
	jpeg_stdio_dest(&compressObj, outfile);

	compressObj.image_width = width;
	compressObj.image_height = height;
	compressObj.input_components = 3;
	compressObj.in_color_space = JCS_RGB;

	jpeg_set_defaults(&compressObj);
	jpeg_set_quality(&compressObj, sJPEGQuality, TRUE);

	// Set up scanlines 
	
	scanlineBuf = new JSAMPLE[width * 3];
	if (!scanlineBuf)
		return(-1);
	rowPtr[0] = scanlineBuf;
	
	// Compress & save

	jpeg_start_compress(&compressObj, TRUE);

	for (i = height - 1; i >= 0; i--)
	{
		Colour	c;

		p = scanlineBuf;
        for (j = 0; j < width; j++)
 		{
 			c = GetPixel(j, i);
			*p++ = Byte(c[0] * 255.0);
			*p++ = Byte(c[1] * 255.0);
			*p++ = Byte(c[2] * 255.0);
        }

		jpeg_write_scanlines(&compressObj, rowPtr, 1);
	}

	jpeg_finish_compress(&compressObj);
	fclose(outfile);
	delete[] scanlineBuf;
}

Int Image::LoadJPEG(const Char *filename)
{
    struct jpeg_decompress_struct	cinfo;
    struct jpeg_error_mgr			jerr;
    JSAMPROW						buffer;
    FILE							*file;
	RGBAPixel						*rgba_buf;
	Int								i, j, k, err = -1;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);

    file = fopen(filename, "rb");
    if (!file) 
	{
		fprintf(stderr, "JPEG: could not open file \"%s\"\n", filename);
		return(err);
    }

    jpeg_stdio_src(&cinfo, file);
    jpeg_read_header(&cinfo, TRUE);

	// decompress options
    cinfo.out_color_space = JCS_RGB;
    cinfo.quantize_colors = FALSE;
    cinfo.do_fancy_upsampling = FALSE;
    cinfo.do_block_smoothing = FALSE;

    jpeg_start_decompress(&cinfo);

    SetSize(cinfo.output_width, cinfo.output_height);

	// JSAMPLE better be a byte.
	Assert(sizeof(JSAMPLE) == 1, "JSAMPLE is not a byte");
	Assert(cinfo.output_components == 3, "JCS_RGB request failed");
    buffer = (JSAMPROW) new Byte[cinfo.output_width * cinfo.output_components];
    rgba_buf = new RGBAPixel[cinfo.output_width];

    if (!buffer || !rgba_buf) 
	{
		delete[] buffer; delete[] rgba_buf;
		fprintf(stderr, "JPEG: out of memory\n");
		goto bye;
    }
	
	k = cinfo.output_height;

    while (cinfo.output_scanline < cinfo.output_height) 
	{
        jpeg_read_scanlines(&cinfo, &buffer, 1);

		for (i = 0, j = 0; i < cinfo.output_width; i++) 
		{
			rgba_buf[i].ch[rgba_R] = buffer[j++];
			rgba_buf[i].ch[rgba_G] = buffer[j++];
			rgba_buf[i].ch[rgba_B] = buffer[j++];			
			rgba_buf[i].ch[rgba_A] = 255;			
		}
		SetRGBASpan(--k, 0, cinfo.output_width, rgba_buf);
    }

    jpeg_finish_decompress(&cinfo);
	err = 0;
	delete[] buffer; 
	delete[] rgba_buf;

bye:
    jpeg_destroy_decompress(&cinfo);

    fclose(file);
	return(err);
}

#endif


// --- RGBAImage methods ------------------------------------------------------

Void RGBAImage::SetSize(Int width, Int height)
{
	delete[] data;
	data = new RGBAPixel[width * height];
	SELF.width = width;
	SELF.height = height;
}

Void RGBAImage::Clear(const Colour &c)
{
	RGBAPixel	*p = data, bc = ColourToRGBA(c);
	Int			i;

	for (i = 0; i < width * height; i++, p++)
		*p = bc;
}

Void RGBAImage::SetPixel(Int x, Int y, const Colour &c)
{
	Assert(x >= 0 && y >= 0 && x < width && y < height,
		   "illegal pixel access.");
	data[x + y * width] = ColourToRGBA(c);
}

Colour RGBAImage::GetPixel(Int x, Int y) const
{
	Assert(x >= 0 && y >= 0 && x < width && y < height,
		   "illegal pixel access.");
	return(RGBAToColour(data[x + y * width]));
}

Void RGBAImage::SetSpan(Int row, Int start, Int length, const Colour *src)
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(RGBAImage::SetSpan) illegal span access.");
	Int				i;
	RGBAPixel		*p = data + row * width + start;
	const Colour	*cp = src;
	
	for (i = 0; i < length; i++, p++, cp++)
		*p = ColourToRGBA(*cp);
}

Void RGBAImage::GetSpan(Int row, Int start, Int length, Colour *dst) const
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(RGBAImage::GetSpan) illegal span access.");
	Int			i;
	RGBAPixel	*p = data + row * width + start;
	Colour		*cp = dst;
	
	for (i = 0; i < length; i++, p++, cp++)
		*cp = RGBAToColour(*p);
}

Void RGBAImage::SetRGBASpan(Int row, Int start, Int length, 
							const RGBAPixel *src) 
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(RGBAImage::SetRGBASpan) illegal span access.");

	memcpy(data + row * width + start, src, length * sizeof(RGBAPixel));
}

Void RGBAImage::GetRGBASpan(Int row, Int start, Int length,
							RGBAPixel *dst) const
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(RGBAImage::GetRGBASpan) illegal span access.");

	memcpy(dst, data + row * width + start, length * sizeof(RGBAPixel));
}


// --- ByteImage methods ------------------------------------------------------

Void ByteImage::SetSize(Int width, Int height)
{
	delete[] data;
	data = new Byte[width * height];
	SELF.width = width;
	SELF.height = height;
}

Void ByteImage::Clear(const Colour &c)
{
	Byte	*p = data, bc = ColourToByte(c);
	Int			i;

	for (i = 0; i < width * height; i++, p++)
		*p = bc;
}

Void ByteImage::SetPixel(Int x, Int y, const Colour &c)
{
	Assert(x >= 0 && y >= 0 && x < width && y < height,
		   "illegal pixel access.");
	data[x + y * width] = ColourToByte(c);
}

Colour ByteImage::GetPixel(Int x, Int y) const
{
	Assert(x >= 0 && y >= 0 && x < width && y < height, 
		   "illegal pixel access.");
	return(ByteToColour(data[x + y * width]));
}

Void ByteImage::SetRealPixel(Int x, Int y, ClrReal r, ImgChannel ch)
{
	Assert(x >= 0 && y >= 0 && x < width && y < height, 
		   "illegal pixel access.");
	if (ch <= chBlue)
		data[x + y * width] = (Byte) (r * 255.0);
}


ClrReal ByteImage::GetRealPixel(Int x, Int y, ImgChannel ch) const
{
	Assert(x >= 0 && y >= 0 && x < width && y < height, 
		   "illegal pixel access.");

	if (ch <= chBlue)
		return(data[x + y * width] / 255.0);
	else
		return(0.0);
}

Void ByteImage::SetSpan(Int row, Int start, Int length, const Colour *src)
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(ByteImage::SetSpan) illegal span access.");
	Int				i;
	Byte			*p = data + row * width + start;
	const Colour	*cp = src;
	
	for (i = 0; i < length; i++, p++, cp++)
		*p = ColourToByte(*cp);
}

Void ByteImage::GetSpan(Int row, Int start, Int length, Colour *dst) const
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(ByteImage::GetSpan) illegal span access.");
	Int			i;
	Byte		*p = data + row * width + start;
	Colour		*cp = dst;
	
	for (i = 0; i < length; i++, p++, cp++)
		*cp = ByteToColour(*p);
}

Void ByteImage::SetByteSpan(Int row, Int start, Int length, 
							const Byte *src, ImgChannel channel) 
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(ByteImage::SetByteSpan) illegal span access.");

	memcpy(data + row * width + start, src, length);
}

Void ByteImage::GetByteSpan(Int row, Int start, Int length,
							Byte *dst, ImgChannel channel) const
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(ByteImage::GetByteSpan) illegal span access.");

	memcpy(dst, data + row * width + start, length);
}


// --- RGBEImage methods ------------------------------------------------------

Void RGBEImage::SetSize(Int width, Int height)
{
	delete[] data;
	data = new RGBEPixel[width * height];
	SELF.width = width;
	SELF.height = height;
}

Void RGBEImage::Clear(const Colour &c)
{
	RGBEPixel	*p = data, bc = ColourToRGBE(c);
	Int			i;

	for (i = 0; i < width * height; i++, p++)
		*p = bc;
}

Void RGBEImage::SetPixel(Int x, Int y, const Colour &c)
{
	Assert(x >= 0 && y >= 0 && x < width && y < height,
		   "illegal pixel access.");
	data[x + y * width] = ColourToRGBE(c);
}

Colour RGBEImage::GetPixel(Int x, Int y) const
{
	Assert(x >= 0 && y >= 0 && x < width && y < height,
		   "illegal pixel access.");
	return(RGBEToColour(data[x + y * width]));
}

Void RGBEImage::SetSpan(Int row, Int start, Int length, const Colour *src)
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(RGBEImage::SetSpan) illegal span access.");
	Int				i;
	RGBEPixel		*p = data + row * width + start;
	const Colour	*cp = src;
	
	for (i = 0; i < length; i++, p++, cp++)
		*p = ColourToRGBE(*cp);
}

Void RGBEImage::GetSpan(Int row, Int start, Int length, Colour *dst) const
{
	Assert(row >= 0 && row < height && start >= 0 && start + length <= width,
		   "(RGBEImage::GetSpan) illegal span access.");
	Int			i;
	RGBEPixel	*p = data + row * width + start;
	Colour		*cp = dst;
	
	for (i = 0; i < length; i++, p++, cp++)
		*cp = RGBEToColour(*p);
}

const Int kRGBE_Excess = 128;

Colour RGBEToColour(RGBEPixel rgbe)
{
	Colour	result;
	Double	f;
	
	if (rgbe.ch[3] == 0)
		return(cBlack);

	f = ldexp(1.0, (Int) rgbe.ch[3] - (kRGBE_Excess + 8));

	result[0] = f * (rgbe.ch[0] + 0.5);
	result[1] = f * (rgbe.ch[1] + 0.5);
	result[2] = f * (rgbe.ch[2] + 0.5);

	return(result);	
}

RGBEPixel ColourToRGBE(const Colour &c)
{
	RGBEPixel	result;
	Int			e;
	Double		d;
	
	d = MaxCmpt(c);
	if (d <= 1e-32)
	{
		*((UInt32*) result.ch) = 0;
	}
	else
	{
		d = frexp(d, &e) * 256.0 / d;
		
		result.ch[0] = (Byte) (c[0] * d);
		result.ch[1] = (Byte) (c[1] * d);
		result.ch[2] = (Byte) (c[2] * d);
		result.ch[3] = e + kRGBE_Excess;
	}
	
	return(result);	
}

RGBEPixel RGBEMult(RGBEPixel rgbe, Double s)
{
	Colour		temp;
	Double		d, f;
	Int			e;
	RGBEPixel	result;
	
	*((UInt32*) result.ch) = 0;

	if (rgbe.ch[3] == 0)
		return(result);

	f = ldexp(1.0, (Int) rgbe.ch[3] - (kRGBE_Excess + 8));

	temp[0] = f * (rgbe.ch[0] + 0.5);
	temp[1] = f * (rgbe.ch[1] + 0.5);
	temp[2] = f * (rgbe.ch[2] + 0.5);

	d = MaxCmpt(temp);
	if (d <= 1e-32)
		return(result);
	else
	{
		d = frexp(d, &e) * 256.0 / d;
		
		result.ch[0] = (Byte) (temp[0] * d);
		result.ch[1] = (Byte) (temp[1] * d);
		result.ch[2] = (Byte) (temp[2] * d);
		result.ch[3] = e + kRGBE_Excess;
	}

	return(result);	
}


// --- ChannelImage methods ---------------------------------------------------



Int RGBEImage::SavePIC(const Char *filename)
{
    Int			i, j;
	FILE		*pic;
	RGBEPixel	*p, *buffer = new RGBEPixel[width];

	if (!buffer)
		return(-1);
	pic = fopen(filename, "w");
	if (!pic)
		return(-1);
	
	fprintf(pic, "#?RADIANCE\n");
	fprintf(pic, "gcl_out\n");
	fprintf(pic, "FORMAT=32-bit_rle_rgbe\n");
	fprintf(pic, "\n");
	fprintf(pic, "-Y %d +X %d\n", width, height);

    for (j = height - 1; j >= 0; j--)
    {
    	p = data + j * width;
		// convert to radiance
    	for (i = 0; i < width; i++)
			buffer[i] = RGBEMult(p[i], 1.0 / 179.0);
    	fwrite(buffer, sizeof(RGBEPixel), width, pic);
    }

	delete[] buffer;
	
    return(fclose(pic));
}

Int RGBEImage::LoadPIC(const Char *filename)
{
    Int			i, j;
	Int			x, y, level;
	FILE		*pic;
	RGBEPixel	*p, *buffer;
	
	pic = fopen(filename, "r");
	if (!pic)
		return(-1);

	if (fscanf(pic, 
			"#?RADIANCE\n"
			"gcl_out\n"
			"FORMAT=32-bit_rle_rgbe\n\n"
			"-Y %d +X %d\n", &x, &y) != 2)
	{
		_Warning("Couldn't read pic file: bad format/not gcl_out\n");
		fclose(pic);
		return(-1);
	}

	SetSize(x, y);

    for (j = height - 1; j >= 0; j--)
    {
    	p = data + j * width;
		fread(p, sizeof(RGBEPixel), width, pic);
	}
	
	return(fclose(pic));
}
