/*
	File:		Plot.cc

	Function:	Multipurpose utility for plotting various useful things,
				including the Utah sun-sky model.

	Author:		Andrew Willmott

	Copyright:	(c) 1999-2000, Andrew Willmott

	Notes:		
*/

#include "XGraphicsSystem.h"
#include "Image.h"
#include "Timer.h"
#include "ArgParse.h"
#include "VecUtil.h"
#include "ColourSystem.h"

#include "SunSky.h"
#include "BRDF.h"

#include <time.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef DEBUG
#define DBG_COUT if (1) cout
#else 
#define DBG_COUT if (0) cout
#endif

struct ImageStats
{
	Double	avgLum;
	Double	avgLogLum;
	Double	minLum;
	Double	maxLum;
	Colour	avgClr;
	Colour	minClr;
	Colour	maxClr;
	Int		numSamples;
};

Void FindImageStats(Image &img, ImageStats &stats)
{		
	Int		x, y;
	Colour	c;
	Double	lum;
	
	stats.numSamples = 0;
	stats.avgLogLum = 0.0;
	stats.avgLum = 0.0;
	stats.minLum = vl_inf;
	stats.maxLum = -vl_inf;
	stats.avgClr.MakeZero();
	stats.minClr.MakeBlock(vl_inf);
	stats.maxClr.MakeBlock(-vl_inf);
	
	for (y = 0; y < img.Height(); y++)
		for (x = 0; x < img.Width(); x++)
		{
			c = img.GetPixel(x, y);
			lum = dot(cRGBToLum, c);

			if (lum > 0.0)
			{
				stats.avgClr += c;
				stats.avgLum += lum;
				stats.avgLogLum += log(lum);
				FindMinCmpts(c, stats.minClr, stats.minClr);
				FindMaxCmpts(c, stats.maxClr, stats.maxClr);
				stats.minLum = Min(lum, stats.minLum);
				stats.maxLum = Max(lum, stats.maxLum);
				stats.numSamples++;
			}
		}

	stats.avgLum /= stats.numSamples;
	stats.avgLogLum /= stats.numSamples;
	stats.avgClr /= stats.numSamples;
}


Void ToneMapImageCamera(Image &img, Image &imgOut)
{		
	Int		x, y;
	Colour	c;
	
	for (y = 0; y < img.Height(); y++)
		for (x = 0; x < img.Width(); x++)
		{
			c = img.GetPixel(x, y) / 60000.0;
			imgOut.SetPixel(x, y, ClipColour(c));
		}
}

Void LinearMapImage(const Image &img, Double start, Double slope, Image &imgOut)
{		
	Int		x, y;
	Colour	c, cStart;

	cStart.MakeBlock(start);
	
	for (y = 0; y < img.Height(); y++)
		for (x = 0; x < img.Width(); x++)
		{
			c = img.GetPixel(x, y);
			c -= cStart;
			c *= slope;
			imgOut.SetPixel(x, y, ClipColour(c));
		}
}

Void ToneMapImage2(Image &img)
{
	Int		x, y;
	Colour	c;
	ClrReal	lum, newLum;

	for (y = 0; y < img.Height(); y++)
		for (x = 0; x < img.Width(); x++)
		{
			c = img.GetPixel(x, y);
			lum = dot(cRGBToLum, c);
			newLum = pow(lum, 1.0 / 3.0);
			img.SetPixel(x, y, c * newLum / lum);
		}
		
	img.Scale(1.0 / img.MaxComponent());
}

// --- PlotPane methods -------------------------------------------------------

class PlotPane : public XBackedPane
{
public:
	Void			Init();
	Void			InitNoWin();
	Void			SetOptions(Int argc, Char **argv);

	Void			HandleKey(Char c);
	Void			TrackMouse(Int sx, Int sy, Int modifiers);
	Void			PrintKeys();
	
	Void			Draw();
	Void			DrawImage();
	Void			DoToneMap();
	
	Void			PaintBRDF(BRDFMtl &material, Coord c);
	Void			PaintSunSky(Coord sun);
	Void			PaintSunSkyWindow(Coord sun);
	Void			PaintCloudySky(Coord sun);

	Int				drawType;
	RGBAImage 		image;
	RGBEImage 		radImage;
	RGBAPixel		*line;
	SpectralCurve 	*specCurve;

	XOffscreenPane	*buffer;
	Bool			dragging;
	Coord			position;
	Bool			first;

	// options
	Int				noFork;
	StrConst		outFile;
	Int				size;
	Float 			latitude;
	Float 			longitude;
	Int				julianDay;
	Float			time;
	Int				zone;
	Float 			turbidity;
	Int				thetaMap;
	Float			gamma;
	Float			lookTheta;
	Float			lookPhi;
	Float			lookSize;
	Bool 			trackSun;
};

Void PlotPane::Init()
{
	image.SetSize(width, height);
	radImage.SetSize(width, height);

	buffer = new XOffscreenPane;
	xgs->CreateOffscreen(buffer, image.Width(), image.Height());
	Attach(buffer);
	image.Clear(cGrey75);
	line = new RGBAPixel[image.Width()];
	dragging = false;
	Draw();
}

Void PlotPane::InitNoWin()
{
	image.SetSize(size, size);
	radImage.SetSize(size, size);

	image.Clear(cGrey75);
	line = new RGBAPixel[image.Width()];
}

static Void PlotPatch(Image &image, const Colour &c)
{
	const Int	size = 16;    
	const Int	offset = 4;    
	Colour  	patchCols[size];       
	Int			i;
	
	for (i = 0; i < size; i++)
		patchCols[i] = c;

	for (i = offset; i < size + offset; i++)
		image.SetSpan(i, offset, size, patchCols);
}

Void PlotPane::PrintKeys()
{
	cout << "0             Phong model" << endl;
	cout << "1             Cook Torrance model" << endl;
	cout << "2             CIE cloudy sky" << endl;
	cout << "3             Utah sun sky model" << endl;
	cout << "4             CIE x y and z matching functions" << endl;
	cout << "5             Chromaticity spectrum" << endl;
	cout << "6             Window of sun sky model" << endl;
	cout << "7             Black body spectrum" << endl;
	cout << "8             CIE chromaticity diagram" << endl;
	cout << "9             Spectrum of file" << endl;
	cout << "<space>       Toggle between CIE chromaticity diagram and spectrum" << endl;
	cout << "d             Save image as plot-xxx.tif" << endl;
	cout << "p             Save image as plot-xxx.pic (Ward's RGBE format)" << endl;
	cout << "<esc>,q       Quit" << endl;
}

Void PlotPane::DrawImage()
{
	Coord c = position;
	Phong brdf0(0.6, 0.4, 15.0, cGreen, cWhite);
		
	if (sqrlen(c) > 1.0)
		c /= len(c) + 1e-16;

	switch (drawType)
	{
	case 0:
		PaintBRDF(brdf0, c);
		break;
	case 1:
//		PaintBRDF(HeBRDF(1.0, 1.45, 0.4, 1.5, 0.0), c);
		PaintBRDF(CookTorrance(0.4, 0.45, 3.5, cBlue), c);
		break;
	case 2:
		PaintCloudySky(c);
		break;
	case 3:
		PaintSunSky(c);
		break;
	case 4:
		if (position[0] < -0.33333)
			kCIE_xbar.MakeImage(image, 2.0);
		else if (position[0] < 0.33333)
			kCIE_ybar.MakeImage(image, 2.0);
		else
			kCIE_zbar.MakeImage(image, 2.0);
		break;
	case 5:
		{
			Colour clr;
			ClrReal x = 0.5 * (position[0] + 1.0);
			ClrReal y = 0.5 * (position[1] + 1.0);

			ChromaticitySpectrum(x, y).MakeImage(image);
			clr = csDisplay.XYZToGamutClip(Colour(x, y, 1.0 - x - y));
			PlotPatch(image, clr);
		}
		break;
	case 6:
		PaintSunSkyWindow(c);
		break;
	case 7:
		{
			BlackBodySpectrum 	bbody(5000.0 * (position[0] + 1.0));
			Colour 				cieClr;

			bbody.MakeImage(image);
			cieClr = bbody.ToCIE_XYZ();
			// convert to chromaticity
			cieClr /= (cieClr[0] + cieClr[1] + cieClr[2]);
			PlotPatch(image, csDisplay.XYZToGamutClip(cieClr));
		}
		break;
	case 8:
		{
			Colour clr;
			static RGBAImage chrImage;
			ClrReal x = 0.5 * (position[0] + 1.0);
			ClrReal y = 0.5 * (position[1] + 1.0);

			if (chrImage.Width() == 0)
			{
				chrImage.SetSize(image.Width(), image.Height());
				DrawChromaticityImage(chrImage);
			}
			
			image.CopyFrom(0, 0, chrImage);
			clr = csDisplay.ChromaToGamut(Chroma(x, y));
			// convert to brightest-possible display colour
			clr /= MaxCmpt(clr);
			PlotPatch(image, clr);
		}
		break;
	case 9:
		if (specCurve)
		{
			Colour	cieClr;
			
			cieClr = specCurve->ToCIE_XYZ();
			cieClr /= cieClr[1];
			specCurve->MakeImage(image, 1 + Int(specCurve->MaxCmpt()));
			PlotPatch(image, csDisplay.XYZToGamutClip(cieClr));
		}
		else
			image.Clear(HSVCol(drand48() * 360.0, 1.0, 1.0));
		break;
	}

	if (gamma > 0.0)
		image.GammaCorrect(gamma);
}

Void PlotPane::Draw()
{
	DrawImage();
	buffer->PutImage(image);
	HandleExpose();
}

Void PlotPane::HandleKey(Char c)
{
	if (c >= '0' && c <= '9')
	{
		drawType = c - '0';
		Draw();
	}
	else if (c == '?' || c == 'h')
		PrintKeys();
	else if (c == 'd')
	{
		FileName outFile;
		outFile.SetPath("plot.tif");
		outFile.MakeUnique();
		image.Save(outFile);
	}
	else if (c == 't')
	{
		cout << "time = " << time << ", turbidity = " << turbidity << endl;
	}
	else if (c == 'i')
	{
		ImageStats stats;

		FindImageStats(radImage, stats);
		cout << "avgLum = " << stats.avgLum << endl;
		cout << "avgLogLum = " << stats.avgLogLum << " ("
			<< exp(stats.avgLogLum) << ")" << endl;
		cout << "minLum = " << stats.minLum << endl;
		cout << "maxLum = " << stats.maxLum << endl;
		cout << "avgClr = " << stats.avgClr << endl;
		cout << "minClr = " << stats.minClr << endl;
		cout << "maxClr = " << stats.maxClr << endl;
		cout << "EV for ISO 100 ~= " << stats.avgLogLum / log(2) - 1 << endl;
	}	
	else if (c == 'p')
	{
		FileName outFile;
		outFile.SetPath("plot.pic");
		outFile.MakeUnique();
		radImage.SavePIC(outFile.GetPath());
	}
	else if (c == ' ')
	{
		if (drawType == 8) 
			drawType = 5;
		else if (drawType == 5) 
			drawType = 8;
		Draw();
	}
	else if (c == 'q')
		xgs->SignalDone();
	else
		XBackedPane::HandleKey(c);
}

Void PlotPane::TrackMouse(Int sx, Int sy, Int modifiers)
{
	Int	 		x, y;
	Coord		start = 2.0 * Coord((sx + 0.5) / width, 1.0 - (sy + 0.5) / height) - vl_1;
	Int			newModifiers;

	GetMouse(x, y, newModifiers);
	position = start;
	dragging = true;

	while (newModifiers & wmButtons)
	{
		position = 2.0 * Coord((x + 0.5) / width, 1.0 - (y + 0.5) / height) - vl_1;
		Draw();
		GetMouse(x, y, newModifiers);
	} 

	dragging = false;
}


// --- Paint methods ----------------------------------------------------------

Void PlotPane::PaintBRDF(BRDFMtl &material, Coord c)
{
	Int			i, j;
	Vector		light, view(vl_z);
	RGBAPixel	greyPixel = ColourToRGBA(cGrey75);
	
	light = Vector(c, sqrt(1.0 - sqrlen(c)));

	for (i = 0; i < image.Height(); i++)
	{
		for (j = 0; j < image.Width(); j++)
		{
			Vector 	normal;
			Coord 	norm2((j + 0.5) / image.Width(), (i + 0.5) / image.Height());
			Colour	hemiColour;
			
			norm2 = 2.0 * norm2 - vl_1;			

			if (sqrlen(norm2) <= 1.0)
			{
				normal = Vector(norm2, sqrt(1.0 - sqrlen(norm2)));
				
				hemiColour = material.BRDF(normal, light, view);
				ClipColour(hemiColour);
				line[j] = ColourToRGBA(hemiColour);
			}
			else
				line[j] = greyPixel;
		}
		image.SetRGBASpan(i, 0, image.Width(), line);
	}
}


// --- Sky stuff --------------------------------------------------------------

GCLReal	CIEOvercastSkyLuminance(GCLReal theta)
{
	return((1.0 + 2.0 * cos(theta)) / 3.0);
}

Void PlotPane::PaintCloudySky(Coord sun)
{
	Int		i, j;
	GCLReal	zenithbr, maxBr = 0.0;
	Colour	skyColour, baseColour;
	Vector	sunDir = vl_z;
	
	baseColour = Colour(0.1, 0.3, 0.5);
	baseColour /= dot(baseColour, cRGBToLum); // normalise to unit luminance

	for (i = 0; i < image.Height(); i++)
		for (j = 0; j < image.Width(); j++)
		{
			Coord sky((j + 0.5) / image.Width(), (i + 0.5) / image.Height());

			sky = 2.0 * sky - vl_1;			

			if (sqrlen(sky) < 1.0)
			{
				GCLReal theta = acos(1.0 - len(sky));

				zenithbr = 8.6 * sunDir[2] + 0.123;
				skyColour = 0.3 * cWhite
						* CIEOvercastSkyLuminance(theta);

				ClipColour(skyColour);
				
				image.SetPixel(j, i, skyColour);
			}
		}
	DBG_COUT << "maxBr = " << maxBr << endl;
}

const Float kPittLong = 80.22;
const Float kPittLat = 40.5;

Void PlotPane::PaintSunSky(Coord c)
{
	if (!first)
	{
		time = (position[0] + 1.0) * 12.0;
		turbidity = 2.0 + (position[1] + 1.0) * 6.0;
	}
	else 
		first = false;
	
	SunSky 			sunSky(latitude, longitude, zone, julianDay, time, turbidity, false);
	Int				i, j;
	Vector			sunDir = sunSky.GetSunPosition();
	GCLReal			sunAngle = sqrt(sunSky.GetSunSolidAngle() / vl_pi);
	Colour			sunColour;

	sunColour = sunSky.GetSunSpectralRadiance().ToCIE_XYZ();
	DBG_COUT << "sun XYZ = " << sunColour << endl;
	sunColour = csDisplay.XYZToGamut(sunColour);
	DBG_COUT << "sun RGB = " << sunColour << endl;

	DBG_COUT << "time = " << time << ", zone = " << zone << ", turbidity = " << turbidity << endl;
	DBG_COUT << "sun angle = " << sunAngle << ", pos = " << sunDir << endl;

	for (i = 0; i < radImage.Height(); i++)
	{
		for (j = 0; j < radImage.Width(); j++)
		{
			Vector 		hemiDir;
			Coord 		hemiPos((j + 0.5) / radImage.Width(), (i + 0.5) / radImage.Height());
			Colour		hemiColour;
			
			hemiPos = 2.0 * hemiPos - vl_1;

			if (sqrlen(hemiPos) <= 1.0)
			{
				// North = Up, East = left, so hemisphere is 'above' viewer:
				// imagine lying on your back looking up at the sky, head oriented
				// towards north

				if (thetaMap)
				{
					// remap to theta-based coords
					GCLReal	r = len(hemiPos);
					hemiPos *= cos(vl_pi * (1 - r) / 2.0) / r;
				}

				hemiDir = Vector(-hemiPos[1], -hemiPos[0], sqrt(1.0 - sqrlen(hemiPos)));

				if (acos(dot(hemiDir, sunDir)) < sunAngle)
				{
					// this is actually a little beside the point: as
					// the sun subtends about 0.5 degrees, at an image
					// size of 400x400 pixels, the sun will only cover a
					// pixel or so.
					hemiColour = sunColour;
				}
				else
				{
					hemiColour = sunSky.GetSkyXYZRadiance(hemiDir);
					hemiColour = csDisplay.XYZToGamut(hemiColour);
				}
							
				radImage.SetPixel(j, i, hemiColour);
			}
			else
				radImage.SetPixel(j, i, cBlack);
		}
	}

	DoToneMap();
}

inline GCLReal AngleBetween(GCLReal thetav, GCLReal phiv, GCLReal theta, GCLReal phi)
{
	GCLReal		cospsi;

	cospsi = sin(thetav) * sin(theta) * cos(phi - phiv) + cos(thetav) * cos(theta);

	if (cospsi > 1.0)
		return 0;
	if (cospsi < -1.0)
		return vl_pi;

	return acos(cospsi);
}

Vector AngleToDir(GCLReal theta, GCLReal phi)
// compass-based system: theta = elevation from horizon, 
// phi = angle clockwise from north (east = 90)
// transforming to SunSky's x = south, y = east, z = up
{
	Vector	v;

	theta = DegsToRads(theta);
	phi = DegsToRads(180.0 - phi);
	
	v[0] = cos(phi) * cos(theta);
	v[1] = sin(phi) * cos(theta);
	v[2] = sin(theta);

	return(v);
}

Void PlotPane::PaintSunSkyWindow(Coord c)
{
	if (!first)
	{
		time = (position[0] + 1.0) * 12.0;
		turbidity = 2.0 + (position[1] + 1.0) * 6.0;
	}
	else 
		first = false;
	
	SunSky 			sunSky(latitude, longitude, zone, julianDay, time, turbidity, false);

	Int				i, j;
	GCLReal			lookMax;
	GCLReal			sunTheta, sunPhi;
	GCLReal			sunAngle = sqrt(sunSky.GetSunSolidAngle() / vl_pi);
	Colour			sunColour;
	Vector			dir, u, v, lookDir, sunDir = sunSky.GetSunPosition();
	
	sunColour = sunSky.GetSunSpectralRadiance().ToCIE_XYZ();
	DBG_COUT << "sun XYZ = " << sunColour << endl;
	sunColour = csDisplay.XYZToGamut(sunColour);
	DBG_COUT << "sun RGB = " << sunColour << endl;
	sunSky.GetSunThetaPhi(sunTheta, sunPhi);

	// convert to our system
	sunTheta = 90.0 - RadsToDegs(sunTheta);
	sunPhi = 180.0 - RadsToDegs(sunPhi);
	
	if (trackSun)
	{
		DBG_COUT << "targetting to sun" << endl;
		lookTheta = sunTheta;
		lookPhi = sunPhi;
		cerr << "sun: theta = " << sunTheta << " phi = " << sunPhi << endl;
	}
	else
		DBG_COUT << "target: " << lookTheta << " " << lookPhi << endl;

	// construct view frame
	lookDir = AngleToDir(lookTheta, lookPhi);
	u = AngleToDir(0, lookPhi + 90.0);
	v = AngleToDir(lookTheta + 90.0, lookPhi);
	lookMax = 2.0 * tan(DegsToRads(lookSize) / 2.0);

	DBG_COUT << "time = " << time << ", zone = " << zone << ", turbidity = " << turbidity << endl;
	DBG_COUT << "sun angle = " << sunAngle << ", pos = " 
		  << sunTheta << ", " << sunPhi << ", dir = " << sunDir << endl;

	for (i = 0; i < radImage.Height(); i++)
	{
		for (j = 0; j < radImage.Width(); j++)
		{
			Coord 		winPos((j + 0.5) / radImage.Width(), (i + 0.5) / radImage.Height());
			Colour		winColour;

			winPos -= Coord(0.5, 0.5);
			winPos *= lookMax;
			
			dir = lookDir + u * winPos[0] + v * winPos[1];
			
			if (dir[2] >= 0.0)
			{
				normalise(dir);
				if (acos(dot(dir, sunDir)) < sunAngle)
				{
					// sunColour = W.cm^-2.Sr^-1
					winColour = sunColour;
				}
				else
				{
					winColour = sunSky.GetSkyXYZRadiance(dir);
					winColour = csDisplay.XYZToGamut(winColour);
				}
			}
			else
				winColour = ClrMix(cGreen, cGrey10, 0.75) * 60000.0;
										
			radImage.SetPixel(j, i, winColour);
		}
	}

	DoToneMap();
}

Void PlotPane::DoToneMap()
{
	ImageStats	stats;
	Double		scale, EV;
	
	FindImageStats(radImage, stats);

	// roughly approximate exposure value for ISO 100 film
	EV = stats.avgLogLum / log(2) - 1;
	// crop arbitrarily to "sensible" values
	EV = Clip(EV, 6.0, 17.0);
	// use this to scale irradiances so average level is 0.18,
	// a la most light meters
	scale = 0.18 * exp(-(EV + 1) * log(2));
	LinearMapImage(radImage, 0.0, scale, image);
}

Void PlotPane::SetOptions(Int argc, Char **argv)
{
	ArgForm		*arg_format;
	Int			help, cie, bb, sky;
	struct tm	*theTime;
	time_t		unixTime;
	Char		*spec = 0;

	thetaMap = -1;

	drawType = 3;
	size = 200;
	first = true;
	specCurve = 0;
	gamma = 2.2;
	latitude = kPittLat;
	longitude = kPittLong;
	lookTheta = -1.0;
	lookPhi = -1.0;
	lookSize = -1.0;
	
	::time(&unixTime);
	theTime = localtime(&unixTime);
	julianDay = theTime->tm_yday;
	time = theTime->tm_hour + theTime->tm_min / 60.0;
	zone = theTime->tm_hour;
	turbidity = 2.5;

	theTime = gmtime(&unixTime);
	// zone's an int, so this will break for non-hour-aligned time zone people.
	zone = theTime->tm_hour - zone;
	if (zone < 0)
		zone += 24;
		
	String	useLine;

	useLine.Printf("Usage: %s [options]", argv[0]);
	
	arg_format = arg_to_form(0,
		"", 
			useLine.CString(),
		"-help", ARG_FLAG(&help),
			"print help",
		"-sky", ARG_FLAG(&sky),
			"Draw utah sun/sky model as flattened hemisphere",
		"-skywin %f [%f %f]", &lookSize, &lookTheta, &lookPhi,
			"Draw a window of the utah sun/sky model: window size [theta, phi]",
		"-cie", ARG_FLAG(&cie),
			"Draw CIE chromaticity diagram",
		"-spectrum %S", &spec,
			"Draw spectrum from given .sd file",
		"-blackbody", ARG_FLAG(&bb),
			"Draw black body spectrum",
		"", "",
		"-o %S", &outFile,
			"don't display window: output to file",
		"-size %d", &size,
			"Set image/window size",
		"-gamma %f", &gamma,
			"set gamma correction [default: 2.2]",
		"-latlong %f %f", &latitude, &longitude,
			"set latitude and longitude",
		"-day %d", &julianDay,
			"set julian day [1-365]",
		"-turb %f", &turbidity,
			"set turbidity [2-30]",
		"-time %f", &time,
			"set time [0-23]",
		"-zone %d", &zone,
			"set time zone [0-23]",
		"-thetaMap", ARG_FLAG(&thetaMap),
			"produce theta-distorted map suitable for texture mapping",
		"-noFork", ARG_FLAG(&noFork),
			"Don't fork into background",
		0);


	if (argc == 1)
	{
		arg_form_print(arg_format);
		cout << endl;
		exit(0);
	}

	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

	if (help)
	{
		arg_form_print(arg_format);
		cout << endl << "Window keys:" << endl;
		PrintKeys();
		exit(0);
	}
	
	DBG_COUT << "delta from GMT: " << theTime->tm_hour << endl;
	DBG_COUT << "day of year: " << julianDay << " time: " << time << " zone: " << zone << endl;

	if (lookSize >= 0.0)
	{
		trackSun = (lookTheta < 0.0);
		drawType = 6;
	}	
	else if (bb)
		drawType = 7;
	else if (cie)
		drawType = 8;
	else if (spec)
	{
		IrregularSpectralCurve *newCurve = new IrregularSpectralCurve;
		cout << "reading " << spec << endl;
		newCurve->ReadSDFile(FileName().SetPath(spec));
		specCurve = newCurve;
		drawType = 9;
	}
	else if (sky)
		drawType = 3;

#ifdef DEBUG
	noFork = true;
#endif
}

Int main(Int argc, char **argv)
{
	XGraphicsSystem			*xgs;
	FileName				file;
	PlotPane				*plotPane;
	FileName				outFile;
		
	csDisplay = csDellPhosphors;

	plotPane = new PlotPane;
	plotPane->SetOptions(argc, argv);

	if (plotPane->outFile)
	{
		plotPane->InitNoWin();
		plotPane->DrawImage();
		outFile.SetPath(plotPane->outFile);
		plotPane->image.Save(outFile);
		return(0);
	}

	if (!plotPane->noFork && fork())
		return(0);

	xgs = new XGraphicsSystem;
	xgs->CreateWindow(plotPane, "plot", plotPane->size, plotPane->size);
	xgs->Run();

	return(0);
}

